#!/usr/bin/env python

"""
Common classes.

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from compiler.ast import AssAttr, Getattr, Name
import compiler.ast
from micropython.data import Attr, Class, Module
from micropython.errors import *

try:
    set
except NameError:
    from sets import Set as set

# Visitors and activities related to node annotations.

class ASTVisitor:

    "A base class for visitors."

    def default(self, node, *args):
        for n in node.getChildNodes():
            self.dispatch(n)

    def dispatch(self, node, *args):

        "Dispatch using 'node', annotating any raised exceptions."

        # Dispatch via a generic visit method.

        try:
            return node.visit(self, *args)

        # Annotate the exception in case of failure.

        except NodeProcessingError, exc:
            if exc.astnode is None:
                exc.astnode = node

                # NOTE: Should perhaps specialise the subclasses appropriately.

                if isinstance(self, (compiler.ast.Class, compiler.ast.Function, compiler.ast.Module)):
                    exc.unit_name = self.unit.full_name()
                else:
                    exc.unit_name = self.full_name()
            raise

    def possible_accessor_types(self, node, defining_users=1):

        """
        Given annotations made during the inspection process, return all possible
        type names and indications of static usage for a 'node' involved in
        attribute access.

        If 'defining_users' is set to a false value, attempt to get the type
        names specifically applicable to the node, rather than retrieving more
        general definition-based type observations.
        """

        target_names = set()

        # Where an attribute could already be detected and where its nature is
        # not that of a general instance or an unresolved name, attempt to
        # identify it.

        if isinstance(node, (AssAttr, Getattr, Name)):

            # Use any explicit attribute annotation.

            if isinstance(node._attr, Attr):
                attr = node._attr
                target_names.add((attr.parent.full_name(), attr.is_static_attribute()))
                return target_names

            # Otherwise, try and use an expression annotation.

            if isinstance(node, (AssAttr, Getattr)):
                expr = node._expr

                if isinstance(expr, Attr) and expr.get_value():
                    expr = expr.get_value()

                if isinstance(expr, Class):
                    attr = expr.all_class_attributes().get(node.attrname)
                elif isinstance(expr, Module):
                    attr = expr.get(node.attrname)
                else:
                    attr = None

                if attr:
                    target_names.add((attr.parent.full_name(), attr.is_static_attribute()))
                    return target_names

            # Otherwise, attempt to employ the attribute usage observations.

            if node._attrusers:

                # Visit each attribute user.

                for user in node._attrusers:

                    # Since users such as branches may not provide type information,
                    # attempt to find defining users.

                    if defining_users:
                        for def_user in user._attrdefs or [user]:
                            for target_name, is_static in def_user._attrtypes.get(node._username, []):
                                target_names.add((target_name, is_static))
                    else:
                        for target_name, is_static in user._attrspecifictypes.get(node._username, []):
                            target_names.add((target_name, is_static))

        return target_names

def used_by_unit(node):

    """
    Return whether the definition made by a 'node' is actually employed by the
    program unit within which it is found.
    """

    return node.unit and node.unit.parent.has_key(node.unit.name)

# Useful data.

operator_functions = {

    # Binary operations.

    "Add" : "add",
    "Bitand" : "and_",
    "Bitor" : "or_",
    "Bitxor" : "xor",
    "Div" : "div",
    "FloorDiv" : "floordiv",
    "LeftShift" : "lshift",
    "Mod" : "mod",
    "Mul" : "mul",
    "Power" : "pow",
    "RightShift" : "rshift",
    "Sub" : "sub",

    # Unary operations.

    "Invert" : "invert",
    "UnaryAdd" : "pos",
    "UnarySub" : "neg",

    # Augmented assignment.

    "+=" : "iadd",
    "-=" : "isub",
    "*=" : "imul",
    "/=" : "idiv",
    "//=" : "ifloordiv",
    "%=" : "imod",
    "**=" : "ipow",
    "<<=" : "ilshift",
    ">>=" : "irshift",
    "&=" : "iand",
    "^=" : "ixor",
    "|=" : "ior",

    # Comparisons.

    "==" : "eq",
    "!=" : "ne",
    "<" : "lt",
    "<=" : "le",
    ">=" : "ge",
    ">" : "gt",

    # Access and slicing.

    "AssSlice" : "setslice",
    "Slice" : "getslice",
    "AssSubscript" : "setitem",
    "Subscript" : "getitem",
    }

# vim: tabstop=4 expandtab shiftwidth=4
